@@ -1,6 +1,5 @@
 import copy
 import json
-import logging
 import os.path
 import re
 import shutil
@@ -9,6 +8,7 @@ from pathlib import Path
 from typing import List, Optional, Type
 
 from gensim.models import KeyedVectors
+from pydantic import BaseModel
 
 from ...errors import StemmingTrainerError
 from ...models import (
@@ -17,12 +17,18 @@ from ...models import (
     get_stem_dict_path_from_model_path,
     save_stem_dict,
 )
-from ...utils import MeasureTime, loader, remove_file_exit_ok, sort_dict_by_values
+from ...utils import (
+    MeasureTime,
+    get_logger,
+    loader,
+    remove_file_exit_ok,
+    sort_dict_by_values,
+)
 from . import StemDict, StemGenerator, reduce_stem_dict
 from .corpus_stemmer import StemCorpusResult, stem_corpus
 from .default_stem_generator import DefaultStemGenerator
 
-log = logging.getLogger(__name__)
+log = get_logger(__name__)
 
 ITER_FOLDER_PATTERN = re.compile(r"iter-\d+")
 
@@ -45,11 +51,11 @@ def get_stemming_trainer_state_path(base_folder: str) -> str:
     return os.path.join(base_folder, "stemming-trainer-state.json")
 
 
-@dataclass
-class StemmingIterationStats:
+class StemmingIterationStats(BaseModel):
     iteration_number: int
     initial_vocab_size: Optional[int] = None
-    time_measures: dict = field(default_factory=dict)
+    time_measures: dict = {}
+    training_params: Optional[dict] = None
     stem_generator_params: Optional[dict] = None
     stem_corpus_result: Optional[StemCorpusResult] = None
     stem_dict: Optional[dict] = None
@@ -58,22 +64,12 @@ class StemmingIterationStats:
     def stem_dict_size(self) -> int:
         return 0 if self.stem_dict is None else len(self.stem_dict)
 
-    def as_dict(self) -> dict:
-        return {
-            "iteration_number": self.iteration_number,
-            "initial_vocab_size": self.initial_vocab_size,
-            "stem_dict_size": self.stem_dict_size,
-            "time_measures": self.time_measures,
-            "stem_generator_params": self.stem_generator_params,
-            "stem_corpus_result": self.stem_corpus_result.as_dict() if self.stem_corpus_result else None,
-            "stem_dict": self.stem_dict,
-        }
-
 
 @dataclass
 class IterationProgram:
     stem_generator: Optional[StemGenerator] = None
-    training_params: Optional[dict] = None
+    training_params: Optional[dict] = field(default_factory=dict)
+    iteration_kwargs: Optional[dict] = field(default_factory=dict)
 
 
 class StemmingIterationTrainer:
@@ -90,16 +86,20 @@ class StemmingIterationTrainer:
         stem_generator: StemGenerator,
         iteration_number: int,
         corpus_folder: str,
-        base_model: KeyedVectors = None,
-        training_params: dict = None,
+        model: Optional[KeyedVectors] = None,
+        remove_words_not_in_model: bool = False,
+        training_params: Optional[dict] = None,
     ):
         self.trainer = trainer
         self.stem_generator = stem_generator
         self.iteration_number = iteration_number
         self.corpus_folder = corpus_folder
-        self.model = base_model
-        self.stats = StemmingIterationStats(self.iteration_number)
+        self.model = model
         self.training_params = training_params or {}
+        self.remove_words_not_in_model = remove_words_not_in_model
+        self.stats = StemmingIterationStats(
+            iteration_number=self.iteration_number, training_params=self.training_params
+        )
 
     @property
     def iteration_folder(self) -> str:
@@ -133,9 +133,12 @@ class StemmingIterationTrainer:
 
     def run_stemming_iteration(self) -> StemmingIterationStats:
         log.info(f"Running stemming iteration number {self.iteration_number}.")
+        self.load_stats()
         with MeasureTime() as mt:
             self.model = self._try_load_trained_model()
-            if not self.model:
+            if self.model:
+                log.info("Found existing model for iteration, skipping training.")
+            else:
                 self.model = self.train_model()
             self.generate_stemmed_corpus()
             log.info(f"Stemming iteration {self.iteration_number} completed.")
@@ -154,6 +157,7 @@ class StemmingIterationTrainer:
             log.warning(f"Failed to copy base corpus into first iteration folder: {e}")
 
     def train_model(self) -> KeyedVectors:
+        log.info(f"Training model for iteration {self.iteration_number}.")
         self._first_iteration_check()
         with MeasureTime() as mt:
             model = self.trainer.train_model_on_corpus(
@@ -174,8 +178,9 @@ class StemmingIterationTrainer:
 
     def generate_stemmed_corpus(self):
         self.stats.initial_vocab_size = len(self.model.key_to_index)
-        self.stats.stem_dict = self.generate_stem_dict()
-        self.save_stats()  # This is mandatory, as we have to save_stats before stem_corpus.
+        if not self.stats.stem_dict:
+            self.stats.stem_dict = self.generate_stem_dict()
+        self.save_stats()  # We have to save stats before stem_corpus for iteration stem dict to be in the state file.
         if len(self.stats.stem_dict) == 0:
             log.info("Stem dict was empty, skipping corpus stemming.")
             return
@@ -185,11 +190,13 @@ class StemmingIterationTrainer:
         with MeasureTime() as mt:
             complete_stem_dict = self.trainer.collect_complete_stem_dict()
         self.stats.time_measures["collect_complete_stem_dict"] = mt.delta
+        allowed_words = set(self.model.key_to_index.keys()) if self.remove_words_not_in_model else None
         with MeasureTime() as mt:
             self.stats.stem_corpus_result = stem_corpus(
                 original_corpus_path=self.iteration_corpus_path,
                 output_corpus_path=self.next_iteration_corpus_path,
                 stem_dict=complete_stem_dict,
+                allowed_words=allowed_words,
             )
         self.stats.time_measures["stem_corpus"] = mt.delta
         if not self.is_first_iteration and os.path.exists(self.next_iteration_corpus_path):
@@ -198,17 +205,25 @@ class StemmingIterationTrainer:
     def save_stats(self):
         stats_file = get_stats_path(self.iteration_folder)
         with open(stats_file, "w") as file:
-            stats_dict = self.stats.as_dict()
+            stats_dict = self.stats.dict()
             serialized = json.dumps(stats_dict, indent=2, ensure_ascii=False)
             file.write(serialized)
 
+    def load_stats(self):
+        stats_file = get_stats_path(self.iteration_folder)
+        try:
+            with open(stats_file) as file:
+                data = json.load(file)
+                self.stats = StemmingIterationStats(**data)
+        except Exception as e:  # noqa
+            pass
+
 
 class StemmingTrainer:
     def __init__(
         self,
         corpus_folder: str,
         completed_iterations: int = 0,
-        latest_model: Optional[KeyedVectors] = None,
         max_iterations: Optional[int] = 10,
         is_fully_stemmed: bool = False,
         min_change_count: int = 0,
@@ -219,7 +234,6 @@ class StemmingTrainer:
     ):
         self.corpus_folder = corpus_folder
         self.completed_iterations = completed_iterations
-        self.latest_model = latest_model  # TODO: Think about that field.
         self.max_iterations = max_iterations
         self.is_fully_stemmed = is_fully_stemmed
         self.min_change_count = min_change_count
@@ -268,6 +282,7 @@ class StemmingTrainer:
 
     def train(self, save_stem_dict_when_done: bool = True):
         log.info("Starting iterations stemmer training...")
+        self.save_state()
         while True:
             if self.max_iterations and self.completed_iterations >= self.max_iterations:
                 log.info(f"Reached {self.completed_iterations} iterations, quitting.")
@@ -289,43 +304,41 @@ class StemmingTrainer:
 
     def run_iteration(self) -> StemmingIterationStats:
         iteration_number = self.completed_iterations + 1
-        stem_generator = self.get_stem_generator(iteration_number=iteration_number)
-        training_params = self.get_training_params(iteration_number=iteration_number)
+        iteration_program = self.get_iteration_program(iteration_number=iteration_number)
+        stem_generator = self.get_stem_generator(iteration_program=iteration_program)
+        training_params = self.get_training_params(iteration_program=iteration_program)
+        iteration_kwargs = iteration_program.iteration_kwargs if iteration_program else {}
         iteration_trainer = StemmingIterationTrainer(
             trainer=self,
             stem_generator=stem_generator,
             iteration_number=iteration_number,
             corpus_folder=self.corpus_folder,
-            base_model=self.latest_model,
             training_params=training_params,
+            **iteration_kwargs,  # type: ignore
         )
         iteration_stats = iteration_trainer.run_stemming_iteration()
         self.completed_iterations += 1
-        self.latest_model = None
         self.save_state()
         return iteration_stats
 
-    def get_training_params(self, iteration_number: int) -> dict:
-        training_params = copy.deepcopy(self.default_training_params)
+    def get_iteration_program(self, iteration_number: int) -> Optional[IterationProgram]:
         try:
-            iteration_program = self.training_program[iteration_number - 1]
+            return self.training_program[iteration_number - 1]
+        except IndexError:
+            return None
+
+    def get_training_params(self, iteration_program: Optional[IterationProgram]) -> dict:
+        training_params = copy.deepcopy(self.default_training_params)
+        if iteration_program:
             iteration_training_params = iteration_program.training_params or {}
             training_params.update(iteration_training_params)
-        except:  # noqa
-            pass
         return training_params
 
-    def get_stem_generator(self, iteration_number: int) -> StemGenerator:
-        stem_generator = None
-        try:
-            iteration_program = self.training_program[iteration_number - 1]
-            stem_generator = iteration_program.stem_generator
-        except IndexError:
-            pass
-        if stem_generator is None:
-            stem_generator_params = self.default_stem_generator_params
-            stem_generator = self.default_stem_generator_class(**stem_generator_params)
-        return stem_generator
+    def get_stem_generator(self, iteration_program: Optional[IterationProgram]) -> StemGenerator:
+        if iteration_program and iteration_program.stem_generator:
+            return iteration_program.stem_generator
+        stem_generator_params = self.default_stem_generator_params
+        return self.default_stem_generator_class(**stem_generator_params)
 
     def save_state(self):
         state_path = get_stemming_trainer_state_path(self.corpus_folder)
