@@ -1,11 +1,11 @@
-import logging
 import os
 
 from gensim.models import KeyedVectors, Word2Vec, fasttext
 
-log = logging.getLogger(__name__)
+from ..utils import get_logger
+from . import Settings
 
-DATA_FOLDER = os.getenv("DATA_FOLDER_PATH", "./data")
+log = get_logger(__name__)
 
 
 def load_kv(kv_path: str) -> KeyedVectors:
@@ -27,4 +27,4 @@ def load_ft_model(model_path: str):
 
 
 def get_path(*path: str) -> str:
-    return os.path.join(DATA_FOLDER, *path)
+    return os.path.join(Settings.data_folder, *path)
